@@ -1,6 +1,7 @@
 """
 """
 
+# pylint: skip-file
 # mypy: ignore-errors
 
 # stdlib
@@ -35,7 +36,12 @@ class Aireps(Reports):
         """Removes PIREPs before updating raw_reports"""
         return [r for r in reports if r.startswith("ARP")]
 
-    def _post_update(self):
+    async def _post_update(self):
+        self.data = []
+        for report in self.raw:
+            parse(report, self.issued)
+
+    def _post_parse(self):
         self.data = []
         for report in self.raw:
             parse(report, self.issued)
