@@ -5,7 +5,9 @@ Contains the core parsing and indent functions of avwx
 # pylint: disable=redefined-builtin
 
 # stdlib
+import re
 import datetime as dt
+import math
 from calendar import monthrange
 from contextlib import suppress
 from copy import copy
@@ -37,13 +39,13 @@ def dedupe(items: Iterable[Any], only_neighbors: bool = False) -> List[Any]:
     return ret
 
 
-def is_unknown(val: str) -> bool:
+def is_unknown(value: str) -> bool:
     """Returns True if val represents and unknown value"""
-    if not isinstance(val, str):
+    if not isinstance(value, str):
         raise TypeError
-    if not val or val.upper() in ("UNKN",):
+    if not value or value.upper() in ("UNKN", "UNK", "UKN"):
         return True
-    for char in val:
+    for char in value:
         if char not in ("/", "X", "."):
             break
     else:
@@ -51,25 +53,26 @@ def is_unknown(val: str) -> bool:
     return False
 
 
-def get_digit_list(alist: List[str], from_index: int) -> Tuple[List[str], List[str]]:
+def get_digit_list(data: List[str], from_index: int) -> Tuple[List[str], List[str]]:
     """Returns a list of items removed from a given list of strings
     that are all digits from 'from_index' until hitting a non-digit item
     """
     ret = []
-    alist.pop(from_index)
-    while len(alist) > from_index and alist[from_index].isdigit():
-        ret.append(alist.pop(from_index))
-    return alist, ret
+    data.pop(from_index)
+    while len(data) > from_index and data[from_index].isdigit():
+        ret.append(data.pop(from_index))
+    return data, ret
 
 
 def unpack_fraction(num: str) -> str:
     """Returns unpacked fraction string 5/2 -> 2 1/2"""
-    nums = [int(n) for n in num.split("/") if n]
-    if len(nums) == 2 and nums[0] > nums[1]:
-        over = nums[0] // nums[1]
-        rem = nums[0] % nums[1]
-        return f"{over} {rem}/{nums[1]}"
-    return num
+    numbers = [int(n) for n in num.split("/") if n]
+    if not (len(numbers) == 2 and numbers[0] > numbers[1]):
+        return num
+    numerator, denominator = numbers
+    over = numerator // denominator
+    rem = numerator % denominator
+    return f"{over} {rem}/{denominator}"
 
 
 def remove_leading_zeros(num: str) -> str:
@@ -117,6 +120,9 @@ def spoken_number(num: str, literal: bool = False) -> str:
 def make_fraction(num: str, repr: str = None, literal: bool = False) -> Fraction:
     """Returns a fraction dataclass for numbers with / in them"""
     num_str, den_str = num.split("/")
+    # 2-1/2 but not -2 1/2
+    if "-" in num_str and not num_str.startswith("-"):
+        num_str = num_str.replace("-", " ")
     denominator = int(den_str)
     # Multiply multi-digit numerator
     if len(num_str) > 1:
@@ -143,11 +149,17 @@ def make_number(
 
     NOTE: Numerators are assumed to have a single digit. Additional are whole numbers
     """
+    # pylint: disable=too-many-branches
     if not num or is_unknown(num):
         return None
     # Check special
     with suppress(KeyError):
-        value, spoken = (special or {}).get(num) or SPECIAL_NUMBERS[num]
+        item = (special or {}).get(num) or SPECIAL_NUMBERS[num]
+        if isinstance(item, tuple):
+            value, spoken = item
+        else:
+            value = item
+            spoken = spoken_number(str(value), literal=literal)
         return Number(repr or num, value, spoken)
     # Check cardinal direction
     if num in CARDINALS:
@@ -163,11 +175,19 @@ def make_number(
         return make_fraction(num, repr, literal)
     # Handle Minus values with errors like 0M04
     if "M" in num:
-        val_str = num.replace("M", "-")
+        val_str = num.replace("MM", "-").replace("M", "-")
         while val_str[0] != "-":
             val_str = val_str[1:]
     else:
         val_str = num
+    # Check value prefixes
+    speak_prefix = ""
+    if val_str.startswith("ABV "):
+        speak_prefix += "above "
+        val_str = val_str[4:]
+    if val_str.startswith("FL"):
+        speak_prefix += "flight level "
+        val_str, literal = val_str[2:], True
     # Create Number
     if not val_str:
         return None
@@ -178,7 +198,8 @@ def make_number(
             value = 0
     else:
         value = int(val_str)
-    return Number(repr or num, value, spoken_number(speak or str(value), literal))
+    spoken = speak_prefix + spoken_number(speak or str(value), literal)
+    return Number(repr or num, value, spoken)
 
 
 def find_first_in_list(txt: str, str_list: List[str]) -> int:
@@ -213,6 +234,47 @@ def is_possible_temp(temp: str) -> bool:
     return True
 
 
+_numeric = Union[int, float]
+
+
+def relative_humidity(
+    temperature: _numeric, dewpoint: _numeric, unit: str = "C"
+) -> float:
+    """Calculates the relative humidity as a 0 to 1 percentage"""
+
+    def saturation(value: Union[int, float]) -> float:
+        """Returns the saturation vapor pressure without the C constant for humidity calc"""
+        return math.exp((17.67 * value) / (243.5 + value))
+
+    if unit == "F":
+        dewpoint = (dewpoint - 32) * 5 / 9
+        temperature = (temperature - 32) * 5 / 9
+    return saturation(dewpoint) / saturation(temperature)
+
+
+# https://aviation.stackexchange.com/questions/47971/how-do-i-calculate-density-altitude-by-hand
+
+
+def pressure_altitude(pressure: float, altitude: _numeric, unit: str = "inHg") -> int:
+    """Calculates the pressure altitude in feet. Converts pressure units"""
+    if unit == "hPa":
+        pressure *= 0.02953
+    return round((29.92 - pressure) * 1000 + altitude)
+
+
+def density_altitude(
+    pressure: float, temperature: _numeric, altitude: _numeric, units: Units
+) -> int:
+    """Calculates the density altitude in feet. Converts pressure and temperature units"""
+    if units.temperature == "F":
+        temperature = (temperature - 32) * 5 / 9
+    if units.altimeter == "hPa":
+        pressure *= 0.02953
+    pressure_alt = pressure_altitude(pressure, altitude)
+    standard = 15 - (2 * altitude / 1000)
+    return round(((temperature - standard) * 120) + pressure_alt)
+
+
 def get_station_and_time(
     data: List[str],
 ) -> Tuple[List[str], Optional[str], Optional[str]]:
@@ -222,12 +284,64 @@ def get_station_and_time(
     station = data.pop(0)
     if not data:
         return data, station, None
-    qtime, rtime = data[0], None
-    if data and qtime.endswith("Z") and qtime[:-1].isdigit():
-        rtime = data.pop(0)
-    elif data and len(qtime) == 6 and qtime.isdigit():
-        rtime = data.pop(0) + "Z"
-    return data, station, rtime
+    q_time, r_time = data[0], None
+    if data and q_time.endswith("Z") and q_time[:-1].isdigit():
+        r_time = data.pop(0)
+    elif data and len(q_time) == 6 and q_time.isdigit():
+        r_time = data.pop(0) + "Z"
+    return data, station, r_time
+
+
+def is_wind(text: str) -> bool:
+    """Returns True if the text is likely a normal wind element"""
+    # Ignore wind shear
+    if text.startswith("WS"):
+        return False
+    # 09010KT, 09010G15KT
+    if len(text) > 4:
+        for ending in ("KT", "KTS", "MPS", "KMH"):
+            if text.endswith(ending):
+                return True
+    # 09010  09010G15 VRB10
+    if not (
+        len(text) == 5
+        or (len(text) >= 8 and text.find("G") != -1)
+        and text.find("/") == -1
+    ):
+        return False
+    return text[:5].isdigit() or (text.startswith("VRB") and text[3:5].isdigit())
+
+
+VARIABLE_DIRECTION_PATTERN = re.compile(r"\d{3}V\d{3}")
+
+
+def is_variable_wind_direction(text: str) -> bool:
+    """Returns True if element looks like 350V040"""
+    if len(text) < 7:
+        return False
+    return VARIABLE_DIRECTION_PATTERN.match(text[:7]) is not None
+
+
+def separate_wind(text: str) -> Tuple[str, str, str]:
+    """Extracts the direction, speed, and gust from a wind element"""
+    direction, speed, gust = "", "", ""
+    # Remove gust
+    if "G" in text:
+        g_index = text.find("G")
+        start, end = g_index + 1, g_index + 3
+        # 16006GP99KT ie gust greater than
+        if "GP" in text:
+            end += 1
+        gust = text[start:end]
+        text = text[:g_index] + text[end:]
+    if text:
+        # 10G18KT
+        if len(text) == 2:
+            speed = text
+        else:
+            direction = text[:3]
+            speed = text[3:]
+    return direction, speed, gust
 
 
 def get_wind(
@@ -247,29 +361,8 @@ def get_wind(
     variable: List[Number] = []
     if data:
         item = copy(data[0])
-        for rep in ["(E)"]:
-            item = item.replace(rep, "")
-        for replacements in (("O", "0"), ("/", ""), ("LKT", "KT"), ("GG", "G")):
-            item = item.replace(*replacements)
-        # 09010KT, 09010G15KT
-        if (
-            item.endswith("KT")
-            or item.endswith("KTS")
-            or item.endswith("MPS")
-            or item.endswith("KMH")
-            or (
-                (
-                    len(item) == 5
-                    or (len(item) >= 8 and item.find("G") != -1)
-                    and item.find("/") == -1
-                )
-                and (
-                    item[:5].isdigit()
-                    or (item.startswith("VRB") and item[3:5].isdigit())
-                )
-            )
-        ):
-            # In order of frequency
+        if is_wind(item):
+            # Select and remove unit in order of frequency
             if item.endswith("KT"):
                 item = item.replace("KT", "")
             elif item.endswith("KTS"):
@@ -280,26 +373,13 @@ def get_wind(
             elif item.endswith("KMH"):
                 units.wind_speed = "km/h"
                 item = item.replace("KMH", "")
-            direction = item[:3]
-            if "G" in item:
-                g_index = item.find("G")
-                gust = item[g_index + 1 :]
-                speed = item[3:g_index]
-            else:
-                speed = item[3:]
-            direction = item[:3]
+            direction, speed, gust = separate_wind(item)
             data.pop(0)
     # Separated Gust
     if data and 1 < len(data[0]) < 4 and data[0][0] == "G" and data[0][1:].isdigit():
         gust = data.pop(0)[1:]
     # Variable Wind Direction
-    if (
-        data
-        and len(data[0]) == 7
-        and data[0][:3].isdigit()
-        and data[0][3] == "V"
-        and data[0][4:].isdigit()
-    ):
+    if data and is_variable_wind_direction(data[0]):
         for item in data.pop(0).split("V"):
             value = make_number(item, speak=item, literal=True)
             if value is not None:
@@ -363,9 +443,11 @@ def sanitize_cloud(cloud: str) -> str:
     if len(cloud) < 4:
         return cloud
     if not cloud[3].isdigit() and cloud[3] not in ("/", "-"):
+        # Bad "O": FEWO03 -> FEW003
         if cloud[3] == "O":
-            cloud = cloud[:3] + "0" + cloud[4:]  # Bad "O": FEWO03 -> FEW003
-        elif cloud[3] != "U":  # Move modifiers to end: BKNC015 -> BKN015C
+            cloud = cloud[:3] + "0" + cloud[4:]
+        # Move modifiers to end: BKNC015 -> BKN015C
+        elif cloud[3] != "U" and cloud[:4] not in ("BASE", "UNKN"):
             cloud = cloud[:3] + cloud[4:] + cloud[3]
     return cloud
 
@@ -377,6 +459,9 @@ def _null_or_int(val: Optional[str]) -> Optional[int]:
     return int(val)
 
 
+_TOP_OFFSETS = ("-TOPS", "-TOP")
+
+
 def make_cloud(cloud: str) -> Cloud:
     """Returns a Cloud dataclass for a cloud string
 
@@ -389,12 +474,19 @@ def make_cloud(cloud: str) -> Cloud:
     modifier: Optional[str] = None
     cloud = sanitize_cloud(cloud).replace("/", "")
     # Separate top
-    topi = cloud.find("-TOP")
-    if topi > -1:
-        top, cloud = cloud[topi + 4 :], cloud[:topi]
+    for target in _TOP_OFFSETS:
+        topi = cloud.find(target)
+        if topi > -1:
+            top, cloud = cloud[topi + len(target) :], cloud[:topi]
+            break
     # Separate type
+    ## BASE027
+    if cloud.startswith("BASES"):
+        cloud = cloud[5:]
+    elif cloud.startswith("BASE"):
+        cloud = cloud[4:]
     ## VV003
-    if cloud.startswith("VV"):
+    elif cloud.startswith("VV"):
         type, cloud = cloud[:2], cloud[2:]
     ## FEW010
     elif len(cloud) >= 3 and cloud[:3] in CLOUD_LIST:
@@ -493,6 +585,7 @@ def parse_date(
 
     This function assumes the given timestamp is within the hour threshold from current date
     """
+    # pylint: disable=too-many-branches
     # Format date string
     date = date.strip("Z")
     if not date.isdigit():
@@ -500,13 +593,13 @@ def parse_date(
     if time_only:
         if len(date) != 4:
             return None
-        ihour = 0
+        index_hour = 0
     else:
         if len(date) == 4:
             date += "00"
         if len(date) != 6:
             return None
-        ihour = 2
+        index_hour = 2
     # Create initial guess
     if target:
         target = dt.datetime(
@@ -515,7 +608,7 @@ def parse_date(
     else:
         target = dt.datetime.now(tz=dt.timezone.utc)
     day = target.day if time_only else int(date[0:2])
-    hour = int(date[ihour : ihour + 2])
+    hour = int(date[index_hour : index_hour + 2])
     # Handle situation where next month has less days than current month
     # Shifted value makes sure that a month shift doesn't happen twice
     shifted = False
@@ -526,7 +619,7 @@ def parse_date(
         guess = target.replace(
             day=day,
             hour=hour % 24,
-            minute=int(date[ihour + 2 : ihour + 4]) % 60,
+            minute=int(date[index_hour + 2 : index_hour + 4]) % 60,
             second=0,
             microsecond=0,
         )
