@@ -1,3 +1,7 @@
+"""
+Station data and search
+"""
+
 from .meta import __LAST_UPDATED__, station_list, uses_na_format, valid_station
 from .station import Station, nearest
 from .search import search
